"""Tests for the package PyNLDAS2."""

from __future__ import annotations

import io

import numpy as np
import pytest
from shapely import Point, Polygon

import pynldas2 as nldas

GEOM = Polygon(
    [[-69.77, 45.07], [-69.31, 45.07], [-69.31, 45.45], [-69.77, 45.45], [-69.77, 45.07]]
)
VAR = ["prcp", "pet"]
DEF_CRS = 4326
ALT_CRS = 3542
COORDS = (-1431147.7928, 318483.4618)
START = "2000-01-01"
END = "2000-01-12"


def assert_close(a: float, b: float, rtol: float = 1e-2) -> bool:
    assert np.isclose(a, b, rtol=rtol).all()


def test_coords():
    clm = nldas.get_bycoords(COORDS, START, END, crs=ALT_CRS, variables=VAR)
    assert_close(clm.prcp.mean(), 0.0051)
    assert_close(clm.pet.mean(), 0.1346)


def test_coords_xr():
    clm = nldas.get_bycoords(COORDS, START, END, crs=ALT_CRS, variables="prcp", to_xarray=True)
    assert_close(clm.prcp.mean(), 0.0051)


def test_geom():
    clm = nldas.get_bygeom(GEOM, START, END, DEF_CRS, VAR)
    assert_close(clm.prcp.mean(), 0.1534)
    assert_close(clm.pet.mean(), 0.0418)


def test_geom_box():
    clm = nldas.get_bygeom(GEOM.bounds, START, END, DEF_CRS, "prcp")
    assert_close(clm.prcp.mean(), 0.1534)


@pytest.mark.jit
def test_snow():
    clm = nldas.get_bycoords(
        (-89.6, 48.3), "2000-01-01", "2000-01-02", crs=DEF_CRS, variables="prcp", snow=True
    )
    assert_close(clm.snow.mean(), 0.0458)
    clm = nldas.get_bygeom(
        Point(-89.6, 48.3).buffer(0.05),
        "2000-01-01",
        "2000-01-02",
        DEF_CRS,
        "prcp",
        snow=True,
    )
    assert_close(clm.snow.mean().compute().item(), 0.04433)


def test_show_versions():
    f = io.StringIO()
    nldas.show_versions(file=f)
    assert "SYS INFO" in f.getvalue()



"""Customized Hydrodata exceptions."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Generator, Sequence


class DownloadError(Exception):
    """Error raised when download fails."""

    def __init__(self, url: str, err_msg: Exception) -> None:
        message = f"Failed to download from {url}:\n{err_msg}"
        super().__init__(message)


class NLDASServiceError(Exception):
    """Exception raised when NLDAS2 web service returns an error."""

    def __init__(self, err: str) -> None:
        self.message = f"NLDAS2 web service returned the following error:\n{err}"
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the error message."""
        return self.message


class InputValueError(Exception):
    """Exception raised for invalid input.

    Parameters
    ----------
    inp : str
        Name of the input parameter
    valid_inputs : tuple
        List of valid inputs
    given : str, optional
        The given input, defaults to None.
    """

    def __init__(
        self,
        inp: str,
        valid_inputs: Sequence[str | int] | Generator[str | int, None, None],
        given: str | int | None = None,
    ) -> None:
        if given is None:
            self.message = f"Given {inp} is invalid. Valid options are:\n"
        else:
            self.message = f"Given {inp} ({given}) is invalid. Valid options are:\n"
        self.message += "\n".join(str(i) for i in valid_inputs)
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the error message."""
        return self.message


class InputTypeError(TypeError):
    """Exception raised when a function argument type is invalid.

    Parameters
    ----------
    arg : str
        Name of the function argument
    valid_type : str
        The valid type of the argument
    example : str, optional
        An example of a valid form of the argument, defaults to None.
    """

    def __init__(self, arg: str, valid_type: str, example: str | None = None) -> None:
        self.message = f"The {arg} argument should be of type {valid_type}"
        if example is not None:
            self.message += f":\n{example}"
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the error message."""
        return self.message


class InputRangeError(ValueError):
    """Exception raised when a function argument is not in the valid range.

    Parameters
    ----------
    variable : str
        Variable with invalid value
    valid_range : str
        Valid range
    """

    def __init__(self, variable: str, valid_range: str) -> None:
        self.message = f"Valid range for {variable} is {valid_range}."
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the error message."""
        return self.message



"""Tests for exceptions and requests."""

from __future__ import annotations

import pytest
from shapely import Polygon

import pynldas2 as nldas
from pynldas2.exceptions import InputRangeError, InputTypeError, InputValueError

GEOM = Polygon(
    [[-69.77, 45.07], [-69.31, 45.07], [-69.31, 45.45], [-69.77, 45.45], [-69.77, 45.07]]
)
VAR = ["prcp", "pet"]
DEF_CRS = 4326
ALT_CRS = 3542
COORDS = (-1431147.7928, 318483.4618)
START = "2000-01-01"
END = "2000-01-12"


def test_coords_range_coords():
    with pytest.raises(InputRangeError) as ex:
        _ = nldas.get_bycoords(COORDS, START, END, crs=DEF_CRS, variables=VAR)
    assert "range" in str(ex.value)


def test_coords_range_start():
    with pytest.raises(InputRangeError) as ex:
        _ = nldas.get_bycoords(COORDS, "1970-01-01", END, crs=ALT_CRS, variables=VAR)
    assert "start_date" in str(ex.value)


def test_coords_range_end():
    with pytest.raises(InputRangeError) as ex:
        _ = nldas.get_bycoords(COORDS, START, "1970-01-01", crs=ALT_CRS, variables=VAR)
    assert "end_date" in str(ex.value)


def test_coords_range_end_before_start():
    with pytest.raises(InputRangeError) as ex:
        _ = nldas.get_bycoords(COORDS, START, "1980-01-01", crs=ALT_CRS, variables=VAR)
    assert "after start_date" in str(ex.value)


def test_coords_type():
    with pytest.raises(InputTypeError) as ex:
        _ = nldas.get_bycoords(COORDS[0], START, END, crs=ALT_CRS, variables=VAR)
    assert "tuple" in str(ex.value)


def test_coords_type_len():
    with pytest.raises(InputTypeError) as ex:
        _ = nldas.get_bycoords(COORDS + COORDS, START, END, crs=ALT_CRS, variables=VAR)
    assert "The coords argument" in str(ex.value)


def test_coords_var():
    with pytest.raises(InputValueError) as ex:
        _ = nldas.get_bycoords(COORDS, START, END, crs=ALT_CRS, variables="tmin")
    assert "prcp" in str(ex.value)



from __future__ import annotations

import asyncio
import contextlib
import sys
from typing import TYPE_CHECKING

import aiofiles
from aiohttp import ClientSession, ClientTimeout, TCPConnector

if TYPE_CHECKING:
    from collections.abc import Sequence
    from pathlib import Path

__all__ = ["stream_write"]
CHUNK_SIZE = 1024 * 1024  # Default chunk size of 1 MB
MAX_HOSTS = 5  # Maximum connections to a single host (rate-limited service)
TIMEOUT = 10 * 60  # Timeout for requests in seconds

if sys.platform == "win32":  # pragma: no cover
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


class ServiceError(Exception):
    """Exception raised for download errors."""

    def __init__(self, err: str, url: str | None = None) -> None:
        self.message = (
            f"Service returned the following error:\nURL: {url}\nERROR: {err}" if url else err
        )
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


async def _stream_file(session: ClientSession, url: str, filepath: Path) -> None:
    """Stream the response to a file, skipping if already downloaded."""
    async with session.get(url) as response:
        if response.status != 200:
            raise ServiceError(await response.text(), str(response.url))
        remote_size = int(response.headers.get("Content-Length", -1))
        if filepath.exists() and filepath.stat().st_size == remote_size:
            return

        async with aiofiles.open(filepath, "wb") as file:
            async for chunk in response.content.iter_chunked(CHUNK_SIZE):
                await file.write(chunk)


async def _stream_session(urls: Sequence[str], files: Sequence[Path]) -> None:
    """Download files concurrently."""
    async with ClientSession(
        connector=TCPConnector(limit_per_host=MAX_HOSTS),
        timeout=ClientTimeout(TIMEOUT),
    ) as session:
        tasks = [_stream_file(session, url, filepath) for url, filepath in zip(urls, files)]
        await asyncio.gather(*tasks)


def is_jupyter_kernel():
    """Check if the code is running in a Jupyter kernel (not IPython terminal)."""
    try:
        from IPython import get_ipython

        ipython = get_ipython()
    except (ImportError, NameError):
        return False
    if ipython is None:
        return False
    return "Terminal" not in ipython.__class__.__name__


def _get_or_create_event_loop() -> tuple[asyncio.AbstractEventLoop, bool]:
    """Create an event loop."""
    try:
        loop = asyncio.get_running_loop()
        new_loop = False
    except RuntimeError:
        loop = asyncio.new_event_loop()
        new_loop = True
    asyncio.set_event_loop(loop)
    if is_jupyter_kernel():
        import nest_asyncio

        nest_asyncio.apply(loop)
    return loop, new_loop


def stream_write(urls: Sequence[str], file_paths: Sequence[Path]) -> None:
    """Download multiple files concurrently by streaming their content to disk."""
    parent_dirs = {filepath.parent for filepath in file_paths}
    for parent_dir in parent_dirs:
        parent_dir.mkdir(parents=True, exist_ok=True)

    loop, is_new_loop = _get_or_create_event_loop()

    try:
        loop.run_until_complete(_stream_session(urls, file_paths))
    finally:
        if is_new_loop:
            loop.close()



"""Get hourly NLDAS2 forcing data."""

from __future__ import annotations

import functools
import hashlib
import itertools
import os
import re
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, TypeVar
from urllib.parse import urlencode

import numpy as np
import numpy.typing as npt
import pandas as pd
import xarray as xr
from pandas.errors import EmptyDataError

import pynldas2._utils as utils
from pynldas2._streaming import stream_write
from pynldas2.exceptions import InputRangeError, InputTypeError, InputValueError, NLDASServiceError

try:
    from numpy._core._exceptions import UFuncTypeError
except ImportError:
    UFuncTypeError = TypeError

try:
    from numba import config as numba_config
    from numba import njit, prange

    ngjit = functools.partial(njit, nogil=True)
    numba_config.THREADING_LAYER = "workqueue"
    has_numba = True
except ImportError:
    has_numba = False
    prange = range

    T = TypeVar("T")
    Func = Callable[..., T]

    def ngjit(_: str | Func[T]) -> Callable[[Func[T]], Func[T]]:
        def decorator_njit(func: Func[T]) -> Func[T]:
            @functools.wraps(func)
            def wrapper_decorator(*args: tuple[Any, ...], **kwargs: dict[str, Any]) -> T:
                return func(*args, **kwargs)

            return wrapper_decorator

        return decorator_njit


if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

    from pyproj import CRS
    from shapely import Polygon

    CRSType = int | str | CRS
    Dataset = TypeVar("Dataset", pd.DataFrame, xr.Dataset)

# Default snow params from https://doi.org/10.5194/gmd-11-1077-2018
T_RAIN = 2.5  # degC
T_SNOW = 0.6  # degC
URL = "https://hydro1.gesdisc.eosdis.nasa.gov/daac-bin/access/timeseries.cgi"

NLDAS2_VARS = {
    "prcp": {"nldas_name": "Rainf", "long_name": "Total precipitation", "units": "kg/m^2"},
    "rlds": {
        "nldas_name": "LWdown",
        "long_name": "Surface downward longwave radiation",
        "units": "W/m^2",
    },
    "rsds": {
        "nldas_name": "SWdown",
        "long_name": "Surface downward shortwave radiation",
        "units": "W/m^2",
    },
    "pet": {"nldas_name": "PotEvap", "long_name": "Potential evaporation", "units": "kg/m^2"},
    "psurf": {"nldas_name": "PSurf", "long_name": "Surface pressure", "units": "Pa"},
    "humidity": {
        "nldas_name": "Qair",
        "long_name": "2-m above ground specific humidity",
        "units": "kg/kg",
    },
    "temp": {
        "nldas_name": "Tair",
        "long_name": "2-m above ground temperature",
        "units": "K",
    },
    "wind_u": {
        "nldas_name": "Wind_E",
        "long_name": "U wind component at 10-m above the surface",
        "units": "m/s",
    },
    "wind_v": {
        "nldas_name": "Wind_N",
        "long_name": "V wind component at 10-m above the surface",
        "units": "m/s",
    },
}
DATE_COL = "Date&Time"
DATE_FMT = "%Y-%m-%dT%H"

__all__ = ["get_bycoords", "get_bygeom"]


@ngjit("f8[::1](f8[::1], f8[::1], f8, f8)")
def _separate_snow(
    prcp: npt.NDArray[np.float64],
    temp: npt.NDArray[np.float64],
    t_rain: np.float64,
    t_snow: np.float64,
) -> npt.NDArray[np.float64]:
    """Separate snow in precipitation."""
    t_rng = t_rain - t_snow
    snow = np.zeros_like(prcp)

    for t in prange(prcp.shape[0]):
        if temp[t] > t_rain:
            snow[t] = 0.0
        elif temp[t] < t_snow:
            snow[t] = prcp[t]
        else:
            snow[t] = prcp[t] * (t_rain - temp[t]) / t_rng
    return snow


def _snow_point(climate: pd.DataFrame, t_rain: float, t_snow: float) -> pd.DataFrame:
    """Separate snow from precipitation."""
    clm = climate.copy()
    clm["snow"] = _separate_snow(
        clm["prcp"].to_numpy("f8"),
        clm["temp"].to_numpy("f8"),
        np.float64(t_rain),
        np.float64(t_snow),
    )
    return clm


def _snow_gridded(climate: xr.Dataset, t_rain: float, t_snow: float) -> xr.Dataset:
    """Separate snow from precipitation."""
    clm = climate.copy()

    def snow_func(
        prcp: npt.NDArray[np.float64],
        temp: npt.NDArray[np.float64],
        t_rain: float,
        t_snow: float,
    ) -> npt.NDArray[np.float64]:
        """Separate snow based on Martinez and Gupta (2010)."""
        return _separate_snow(
            prcp.astype("f8"),
            temp.astype("f8"),
            np.float64(t_rain),
            np.float64(t_snow),
        )

    clm["snow"] = xr.apply_ufunc(
        snow_func,
        clm["prcp"],
        clm["temp"],
        t_rain,
        t_snow,
        input_core_dims=[["time"], ["time"], [], []],
        output_core_dims=[["time"]],
        vectorize=True,
        output_dtypes=[clm["prcp"].dtype],
    ).transpose("time", "y", "x")
    clm["snow"].attrs["units"] = "mm"
    clm["snow"].attrs["long_name"] = "Snowfall hourly total"
    return clm


def separate_snow(clm: Dataset, t_rain: float = T_RAIN, t_snow: float = T_SNOW) -> Dataset:
    """Separate snow based on :footcite:t:`Martinez_2010`.

    Parameters
    ----------
    clm : pandas.DataFrame or xarray.Dataset
        Climate data that should include ``prcp`` and ``temp``.
    t_rain : float, optional
        Threshold for temperature in deg C for considering rain, defaults to 2.5 degrees C.
    t_snow : float, optional
        Threshold for temperature in deg C for considering snow, defaults to 0.6 degrees C.

    Returns
    -------
    pandas.DataFrame or xarray.Dataset
        Input data with ``snow`` column if input is a ``pandas.DataFrame``,
        or ``snow`` variable if input is an ``xarray.Dataset``.

    References
    ----------
    .. footbibliography::
    """
    if not has_numba:
        warnings.warn(
            "Numba not installed. Using slow pure python version.", UserWarning, stacklevel=2
        )

    if not isinstance(clm, (pd.DataFrame, xr.Dataset)):
        raise InputTypeError("clm", "pandas.DataFrame or xarray.Dataset")

    if isinstance(clm, xr.Dataset):
        return _snow_gridded(clm, t_rain + 273.15, t_snow + 273.15)
    return _snow_point(clm, t_rain + 273.15, t_snow + 273.15)


def _txt2df(name: str, txt_file: Path) -> pd.Series:
    """Convert text to dataframe."""
    try:
        data = pd.read_csv(txt_file, skiprows=12, sep=r"\s+").dropna()
        data.index = pd.to_datetime(data[DATE_COL], utc=True)
    except EmptyDataError:
        return pd.Series(name=name)
    except UFuncTypeError as ex:
        msg = "".join(re.findall("<strong>(.*?)</strong>", txt_file.read_text(), re.DOTALL)).strip()
        raise NLDASServiceError(msg) from ex

    data = data["Data"]
    data.name = name
    data.index.name = "time"
    data.index = pd.to_datetime(data.index).tz_localize(None)
    return data


def _get_variables(
    variables: str | list[str] | None, snow: bool
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    """Get variables."""
    source_tag = "NLDAS2:NLDAS_FORA0125_H_v2.0"
    if variables is None:
        url_vars = tuple(f"{source_tag}:{d['nldas_name']}" for d in NLDAS2_VARS.values())
        clm_vars = tuple(NLDAS2_VARS)
    else:
        url_vars = [variables] if isinstance(variables, str) else list(variables)
        if snow:
            url_vars = list(set(url_vars).union({"temp", "prcp"}))
        if any(v not in NLDAS2_VARS for v in url_vars):
            raise InputValueError("variables", list(NLDAS2_VARS))
        url_vars, clm_vars = zip(
            *((f"{source_tag}:{NLDAS2_VARS[v]['nldas_name']}", v) for v in url_vars)
        )
    return url_vars, clm_vars


def _get_dates(
    start_date: str,
    end_date: str,
) -> list[pd.Timestamp]:
    """Get dates."""
    start = pd.to_datetime(start_date)
    end = pd.to_datetime(end_date) + pd.Timedelta("1D")
    if start < pd.to_datetime("1979-01-01T13"):
        raise InputRangeError("start_date", "1979-01-01 to yesterday")
    if end > pd.Timestamp.now() - pd.Timedelta("1D"):
        raise InputRangeError("end_date", "1979-01-01 to yesterday")
    if end <= start:
        raise InputRangeError("end_date", "after start_date")

    dates = pd.date_range(start, end, freq="10000D").tolist()
    dates = [*dates, end] if dates[-1] < end else dates
    return dates


def _download_files(
    lons: Iterable[float],
    lats: Iterable[float],
    variables: str | list[str] | None,
    snow: bool,
    start_date: str,
    end_date: str,
) -> dict[tuple[float, float], dict[str, list[Path]]]:
    """Download NLDAS data and return a dictionary of files grouped by location and variable."""
    dates = _get_dates(start_date, end_date)
    url_vars, clm_vars = _get_variables(variables, snow)
    meta, urls = zip(
        *(
            (
                (float(lon), float(lat), v),
                f"{URL}?"
                + urlencode(
                    {
                        "variable": url_v,
                        "location": f"GEOM:POINT({lon}, {lat})",
                        "startDate": s.strftime(DATE_FMT),
                        "endDate": e.strftime(DATE_FMT),
                        "type": "asc2",
                    }
                ),
            )
            for lon, lat in zip(lons, lats)
            for (url_v, v), (s, e) in itertools.product(
                zip(url_vars, clm_vars), zip(dates[:-1], dates[1:])
            )
        )
    )

    hr_cache = os.getenv("HYRIVER_CACHE_NAME")
    cache_dir = Path(hr_cache).parent if hr_cache else Path("cache")
    cache_dir.mkdir(exist_ok=True, parents=True)
    file_paths = [
        cache_dir / f"{x}_{y}_{v}_{hashlib.sha256(url.encode()).hexdigest()}.txt"
        for (x, y, v), url in zip(meta, urls)
    ]
    stream_write(urls, file_paths)
    # group based on lon, lat, and variable, i.e, dict of dict of list
    grouped_files = {}
    for (x, y, v), f in zip(meta, file_paths):
        if (x, y) not in grouped_files:
            grouped_files[(x, y)] = {}
        if v not in grouped_files[(x, y)]:
            grouped_files[(x, y)][v] = []

        grouped_files[(x, y)][v].append(f)
    return grouped_files


def _by_coord(
    txt_files: dict[str, list[Path]],
    start_date: str,
    end_date: str,
    snow: bool,
    snow_params: dict[str, float] | None,
) -> pd.DataFrame:
    """Get NLDAS climate forcing data for a single location."""
    clm = (
        pd.concat((pd.concat(_txt2df(v, f) for f in resp) for v, resp in txt_files.items()), axis=1)
        .loc[start_date:end_date]
        .rename(columns={d["nldas_name"]: n for n, d in NLDAS2_VARS.items()})
    )
    if snow:
        params = {"t_rain": T_RAIN, "t_snow": T_SNOW} if snow_params is None else snow_params
        clm = separate_snow(clm, **params)
    return clm


def _get_lon_lat(
    coords: list[tuple[float, float]] | tuple[float, float],
    bounds: tuple[float, float, float, float],
    coords_id: Sequence[str | int] | None,
    crs: CRSType,
    to_xarray: bool,
) -> tuple[list[float], list[float]]:
    """Get longitude and latitude from a list of coordinates."""
    coords_list = utils.transform_coords(coords, crs, 4326)

    if to_xarray and coords_id is not None and len(coords_id) != len(coords_list):
        raise InputTypeError("coords_id", "list with the same length as of coords")

    lon, lat = utils.validate_coords(coords_list, bounds).T
    return lon.tolist(), lat.tolist()


def get_bycoords(
    coords: list[tuple[float, float]],
    start_date: str,
    end_date: str,
    coords_id: Sequence[str | int] | None = None,
    crs: CRSType = 4326,
    variables: str | list[str] | None = None,
    to_xarray: bool = False,
    snow: bool = False,
    snow_params: dict[str, float] | None = None,
) -> pd.DataFrame | xr.Dataset:
    """Get NLDAS-2 climate forcing data for a list of coordinates.

    Parameters
    ----------
    coords : list of tuples
        List of (lon, lat) coordinates.
    start_date : str
        Start date of the data.
    end_date : str
        End date of the data.
    crs : str, int, or pyproj.CRS, optional
        The CRS of the input coordinates, defaults to ``EPSG:4326``.
    variables : str or list of str, optional
        Variables to download. If None, all variables are downloaded.
        Valid variables are: ``prcp``, ``pet``, ``temp``, ``wind_u``, ``wind_v``,
        ``rlds``, ``rsds``, and ``humidity`` and ``psurf``.
    to_xarray : bool, optional
        If True, the data is returned as an xarray dataset.
    snow : bool, optional
        Compute snowfall from precipitation and temperature. Defaults to ``False``.
    snow_params : dict, optional
        Model-specific parameters as a dictionary that is passed to the snowfall function.
        These parameters are only used if ``snow`` is ``True``. Two parameters are required:
        ``t_rain`` (deg C) which is the threshold for temperature for considering rain and
        ``t_snow`` (deg C) which is the threshold for temperature for considering snow.
        The default values are ``{'t_rain': 2.5, 't_snow': 0.6}`` that are adopted from
        https://doi.org/10.5194/gmd-11-1077-2018.

    Returns
    -------
    pandas.DataFrame
        The requested data as a dataframe.
    """
    bounds = (-125.0, 25.0, -67.0, 53.0)
    lons, lats = _get_lon_lat(coords, bounds, coords_id, crs, to_xarray)
    n_pts = len(lons)
    grouped_files = _download_files(lons, lats, variables, snow, start_date, end_date)

    idx = list(coords_id) if coords_id is not None else list(range(n_pts))
    idx = dict(zip(zip(lons, lats), idx))
    clm_dict = {
        idx[c]: _by_coord(file, start_date, end_date, snow, snow_params)
        for c, file in grouped_files.items()
    }
    if to_xarray:
        clm_ds = xr.concat(
            (xr.Dataset.from_dataframe(clm) for clm in clm_dict.values()),
            dim=pd.Index(list(clm_dict), name="id"),
        )
        clm_ds.attrs["tz"] = "UTC"
        return clm_ds

    if n_pts == 1:
        clm = next(iter(clm_dict.values()), pd.DataFrame())
    else:
        clm = pd.concat(clm_dict.values(), keys=list(clm_dict), axis=1)
        clm.columns = clm.columns.set_names(["id", "variable"])
    clm.index = pd.DatetimeIndex(clm.index, tz="UTC")
    clm.index.name = "time"
    return clm


def _txt2da(
    lon: float,
    lat: float,
    name: str,
    txt_file: Path,
) -> xr.DataArray:
    """Convert text to dataarray."""
    data = _txt2df(name, txt_file)
    da = data.to_xarray()
    return da.assign_coords(x=lon, y=lat).expand_dims("y").expand_dims("x")


def get_bygeom(
    geometry: Polygon | tuple[float, float, float, float],
    start_date: str,
    end_date: str,
    geo_crs: CRSType = 4326,
    variables: str | list[str] | None = None,
    snow: bool = False,
    snow_params: dict[str, float] | None = None,
) -> xr.Dataset:
    """Get hourly NLDAS-2 climate forcing within a geometry at 0.125 resolution.

    Parameters
    ----------
    geometry : Polygon or tuple
        The geometry of the region of interest. It can be a shapely Polygon or a tuple
        of length 4 representing the bounding box (minx, miny, maxx, maxy).
    start_date : str
        Start date of the data.
    end_date : str
        End date of the data.
    geo_crs : int, str, or pyproj.CRS
        CRS of the input geometry
    variables : str or list of str, optional
        Variables to download. If None, all variables are downloaded.
        Valid variables are: ``prcp``, ``pet``, ``temp``, ``wind_u``, ``wind_v``,
        ``rlds``, ``rsds``, and ``humidity`` and ``psurf``.
    snow : bool, optional
        Compute snowfall from precipitation and temperature. Defaults to ``False``.
    snow_params : dict, optional
        Model-specific parameters as a dictionary that is passed to the snowfall function.
        These parameters are only used if ``snow`` is ``True``. Two parameters are required:
        ``t_rain`` (deg C) which is the threshold for temperature for considering rain and
        ``t_snow`` (deg C) which is the threshold for temperature for considering snow.
        The default values are ``{'t_rain': 2.5, 't_snow': 0.6}`` that are adopted from
        https://doi.org/10.5194/gmd-11-1077-2018.

    Returns
    -------
    xarray.Dataset
        The requested forcing data.
    """
    nldas_grid = utils.get_grid_mask()
    geom = utils.to_geometry(geometry, geo_crs, nldas_grid.rio.crs)
    msk = nldas_grid.CONUS_mask.rio.clip([geom], all_touched=True)
    lons, lats = zip(*itertools.product(msk.get_index("lon"), msk.get_index("lat")))
    grouped_files = _download_files(lons, lats, variables, snow, start_date, end_date)

    clm = xr.merge(
        _txt2da(x, y, f, r)
        for (x, y), v_files in grouped_files.items()
        for f, resp in v_files.items()
        for r in resp
    )
    clm = (
        clm.rename(
            {d["nldas_name"]: n for n, d in NLDAS2_VARS.items() if d["nldas_name"] in clm.data_vars}
        )
        .sel(time=slice(start_date, end_date))
        .transpose("time", "y", "x")
    )
    clm.attrs["tz"] = "UTC"
    for v in clm.data_vars:
        clm[v].attrs = NLDAS2_VARS[str(v)]
    clm = clm.rio.write_crs(nldas_grid.rio.crs)
    if snow:
        params = {"t_rain": T_RAIN, "t_snow": T_SNOW} if snow_params is None else snow_params
        clm = separate_snow(clm, **params)
    return clm



"""Some utilities for PyDaymet."""

from __future__ import annotations

import os
from collections.abc import Iterable, Sequence
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING
from urllib.error import HTTPError
from urllib.request import urlretrieve

import numpy as np
import pyproj
import shapely
import xarray as xr
from pyproj import Transformer
from pyproj.exceptions import CRSError as ProjCRSError
from rioxarray.exceptions import OneDimensionalRaster
from shapely import Polygon, ops

from pynldas2.exceptions import DownloadError, InputRangeError, InputTypeError

if TYPE_CHECKING:
    from numpy.typing import NDArray

    CRSType = int | str | pyproj.CRS
    PolyType = Polygon | tuple[float, float, float, float]
    Number = int | float | np.number

__all__ = [
    "clip_dataset",
    "get_grid_mask",
    "to_geometry",
    "transform_coords",
    "validate_coords",
    "validate_crs",
]

TransformerFromCRS = lru_cache(Transformer.from_crs)


def validate_crs(crs: CRSType) -> str:
    """Validate a CRS.

    Parameters
    ----------
    crs : str, int, or pyproj.CRS
        Input CRS.

    Returns
    -------
    str
        Validated CRS as a string.
    """
    try:
        return pyproj.CRS(crs).to_string()
    except ProjCRSError as ex:
        raise InputTypeError("crs", "a valid CRS") from ex


def transform_coords(
    coords: Sequence[tuple[float, float]], in_crs: CRSType, out_crs: CRSType
) -> list[tuple[float, float]]:
    """Transform coordinates from one CRS to another."""
    try:
        pts = shapely.points(np.atleast_2d(coords))
    except (TypeError, AttributeError, ValueError) as ex:
        raise InputTypeError("coords", "a list of tuples") from ex
    x, y = shapely.get_coordinates(pts).T
    x_proj, y_proj = TransformerFromCRS(in_crs, out_crs, always_xy=True).transform(x, y)
    return list(zip(x_proj, y_proj))


def _geo_transform(geom: Polygon, in_crs: CRSType, out_crs: CRSType) -> Polygon:
    """Transform a geometry from one CRS to another."""
    project = TransformerFromCRS(in_crs, out_crs, always_xy=True).transform
    return ops.transform(project, geom)


def validate_coords(
    coords: Iterable[tuple[float, float]], bounds: tuple[float, float, float, float]
) -> NDArray[np.float64]:
    """Validate coordinates within a bounding box."""
    try:
        pts = shapely.points(list(coords))
    except (TypeError, AttributeError, ValueError) as ex:
        raise InputTypeError("coords", "a list of tuples") from ex
    if shapely.contains(shapely.box(*bounds), pts).all():
        return shapely.get_coordinates(pts).round(6)
    raise InputRangeError("coords", f"within {bounds}")


def to_geometry(
    geometry: Polygon | tuple[float, float, float, float],
    geo_crs: CRSType | None = None,
    crs: CRSType | None = None,
) -> Polygon:
    """Return a Shapely geometry and optionally transformed to a new CRS.

    Parameters
    ----------
    geometry : shaple.Geometry or tuple of length 4
        Any shapely geometry object or a bounding box (minx, miny, maxx, maxy).
    geo_crs : int, str, or pyproj.CRS, optional
        Spatial reference of the input geometry, defaults to ``None``.
    crs : int, str, or pyproj.CRS
        Target spatial reference, defaults to ``None``.

    Returns
    -------
    shapely.Polygon
        A shapely geometry object.
    """
    is_geom = np.atleast_1d(shapely.is_geometry(geometry))
    if is_geom.all() and len(is_geom) == 1:
        geom = geometry
    elif isinstance(geometry, Iterable) and len(geometry) == 4 and np.isfinite(geometry).all():
        geom = shapely.box(*geometry)
    else:
        raise InputTypeError("geometry", "a shapley geometry or tuple of length 4")

    if geo_crs is not None and crs is not None:
        return _geo_transform(geom, geo_crs, crs)
    elif geo_crs is None and crs is not None:
        return geom
    raise InputTypeError("geo_crs/crs", "either both None or both valid CRS")


def clip_dataset(
    ds: xr.Dataset,
    geometry: Polygon | tuple[float, float, float, float],
    crs: CRSType,
) -> xr.Dataset:
    """Mask a ``xarray.Dataset`` based on a geometry."""
    attrs = {v: ds[v].attrs for v in ds}

    geom = to_geometry(geometry, crs, ds.rio.crs)
    try:
        ds = ds.rio.clip_box(*geom.bounds, auto_expand=True)
        if isinstance(geometry, Polygon):
            ds = ds.rio.clip([geom], all_touched=True)
    except OneDimensionalRaster:
        ds = ds.rio.clip([geom], all_touched=True)

    _ = [ds[v].rio.update_attrs(attrs[v], inplace=True) for v in ds]
    ds.rio.update_encoding(ds.encoding, inplace=True)
    return ds


def get_grid_mask(save_dir: str | Path | None = None) -> xr.Dataset:
    """Get the NLDAS-2 grid that contains the land/water/soil/vegetation mask.

    Returns
    -------
    xarray.Dataset
        The grid mask.
    """
    url = "/".join(
        (
            "https://ldas.gsfc.nasa.gov/sites/default",
            "files/ldas/nldas/NLDAS_masks-veg-soil.nc4",
        )
    )
    if save_dir is None:
        hr_cache = os.getenv("HYRIVER_CACHE_NAME")
        save_dir = Path(hr_cache).parent if hr_cache else Path("cache")
    else:
        save_dir = Path(save_dir)
    save_dir.mkdir(exist_ok=True, parents=True)
    nc_path = save_dir / Path(url).name
    if nc_path.exists():
        return xr.open_dataset(nc_path, decode_coords="all")
    try:
        urlretrieve(url, nc_path)
    except HTTPError as ex:
        raise DownloadError(url, str(ex)) from ex
    with xr.open_dataset(nc_path) as ds:
        grid = ds.rio.write_crs(4326).load()
    nc_path.unlink()
    grid.to_netcdf(nc_path)
    return grid



"""Configuration for pytest."""

from __future__ import annotations

import pytest


@pytest.fixture(autouse=True)
def _add_standard_imports(doctest_namespace):
    """Add async_retriever namespace for doctest."""
    import pynldas2 as nldas

    doctest_namespace["nldas"] = nldas



"""Top-level package."""

from __future__ import annotations

from importlib.metadata import PackageNotFoundError, version

from pynldas2 import exceptions
from pynldas2._utils import get_grid_mask
from pynldas2.print_versions import show_versions
from pynldas2.pynldas2 import get_bycoords, get_bygeom

try:
    __version__ = version("pynldas2")
except PackageNotFoundError:
    __version__ = "999"

__all__ = [
    "__version__",
    "exceptions",
    "get_bycoords",
    "get_bygeom",
    "get_grid_mask",
    "show_versions",
]



"""Nox sessions."""

from __future__ import annotations

import shutil
from pathlib import Path
from typing import Literal

import nox

try:
    import tomllib as tomli
except ImportError:
    import tomli


def get_package_name() -> str:
    """Get the name of the package."""
    with Path("pyproject.toml").open("rb") as f:
        return tomli.load(f)["project"]["name"]


def get_extras() -> list[str]:
    """Get the name of the package."""
    with Path("pyproject.toml").open("rb") as f:
        extras = tomli.load(f)["project"]["optional-dependencies"]
    return [e for e in extras if e not in ("test", "typeguard")]


def get_deps() -> list[str]:
    """Get the name of the package."""
    with Path("pyproject.toml").open("rb") as f:
        return tomli.load(f)["project"]["dependencies"]


py39 = ["3.9"]
py312 = ["3.12"]
py313 = ["3.13"]
package = get_package_name()
gh_deps = {
    "async-retriever": [],
    "hydrosignatures": [],
    "pygeoogc": ["async-retriever"],
    "pygeoutils": ["async-retriever", "pygeoogc"],
    "pynhd": ["async-retriever", "pygeoogc", "pygeoutils"],
    "py3dep": ["async-retriever", "pygeoogc", "pygeoutils"],
    "pygeohydro": ["async-retriever", "pygeoogc", "pygeoutils", "pynhd", "hydrosignatures"],
    "pydaymet": [],
    "pygridmet": [],
    "pynldas2": [],
}
nox.options.sessions = (
    "pre-commit",
    "type-check",
    "test39",
    "test312",
)


def install_deps(
    session: nox.Session, extra: str | None = None, version_limit: list[str] | None = None
) -> None:
    """Install package dependencies."""
    deps = [f".[{extra}]"] if extra else ["."]
    deps += [f"git+https://github.com/hyriver/{p}.git" for p in gh_deps[package]]
    if version_limit:
        deps += list(version_limit)
    session.install(*deps)
    dirs = [".pytest_cache", "build", "dist", ".eggs"]
    for d in dirs:
        shutil.rmtree(d, ignore_errors=True)

    patterns = ["*.egg-info", "*.egg", "*.pyc", "*~", "**/__pycache__"]
    for p in patterns:
        for f in Path.cwd().rglob(p):
            shutil.rmtree(f, ignore_errors=True)


@nox.session(name="pre-commit", python=py313, venv_backend="micromamba")
def pre_commit(session: nox.Session) -> None:
    """Lint using pre-commit."""
    session.install("pre-commit")
    session.run(
        "pre-commit",
        "run",
        "--all-files",
        "--hook-stage=manual",
        *session.posargs,
    )


@nox.session(name="pc-update", python=py313, venv_backend="micromamba")
def pc_update(session: nox.Session) -> None:
    """Lint using pre-commit."""
    session.install("pre-commit")
    session.run(
        "pre-commit",
        "autoupdate",
        *session.posargs,
    )


@nox.session(name="type-check", python=py312, venv_backend="micromamba")
def type_check(session: nox.Session) -> None:
    """Run Pyright."""
    extras = get_extras()
    install_deps(session, ",".join(extras))
    session.install("pyright")
    session.run("pyright")


def setup_session(session: nox.Session, with_jit: bool) -> bool:
    """Set up session environment with conda and dependencies."""
    session.conda_install("gdal", channel="conda-forge")
    extras = get_extras()
    jit_dep = "jit" in extras
    if jit_dep and not with_jit:
        extras.remove("jit")
    install_deps(session, ",".join(["test", *extras]))
    return jit_dep


def run_tests(
    session: nox.Session,
    jit: bool,
    py_version: Literal[39, 312],
    extra_args: list[str] | None = None,
) -> None:
    """Run the test suite with optional jit and extra arguments."""
    session.run(
        "pytest",
        "--cov",
        "--cov-append",
        "--cov-branch",
        "--cov-report=xml",
        "--junitxml=junit.xml",
        *(extra_args or []),
        *session.posargs,
    )
    session.notify("cover")
    if jit:
        session.notify(f"jit{py_version}")


@nox.session(python="3.9", venv_backend="micromamba")
def test39(session: nox.Session) -> None:
    """Run the test suite for Python 3.9."""
    jit_dep = setup_session(session, False)
    run_tests(session, jit_dep, 39)


@nox.session(python="3.12", venv_backend="micromamba")
def test312(session: nox.Session) -> None:
    """Run the test suite for Python 3.12."""
    jit_dep = setup_session(session, False)
    run_tests(session, jit_dep, 312)


@nox.session(python="3.9", venv_backend="micromamba")
def jit39(session: nox.Session) -> None:
    """Run tests that require jit dependencies for Python 3.9."""
    setup_session(session, True)
    session.run("pytest", "-m", "jit", *session.posargs)


@nox.session(python="3.12", venv_backend="micromamba")
def jit312(session: nox.Session) -> None:
    """Run tests that require jit dependencies for Python 3.12."""
    setup_session(session, True)
    session.run("pytest", "-m", "jit", *session.posargs)


@nox.session(python=py313, venv_backend="micromamba")
def cover(session: nox.Session) -> None:
    """Coverage analysis."""
    session.install("coverage[toml]")
    session.run("coverage", "report")
    session.run("coverage", "html")
